import java.awt.*;
import Utilities.*;
import Synchronization.*;
import XtangoAnimation.*;

class AnimatedBoundedBuffer extends MyObject { // designed for multiple
                      // producer threads and multiple consumer threads
   private int numSlots = 0;
   private XtangoAnimator xa = null;
   private double[] buffer = null;
   private int putIn = 0, takeOut = 0;
   private int count = 0;
   private BinarySemaphore mutex = null;
   private BinarySemaphore mutexP = null;
   private BinarySemaphore mutexC = null;
   private CountingSemaphore elements = null;
   private CountingSemaphore spaces = null;

   public AnimatedBoundedBuffer(int numSlots, XtangoAnimator xa) {
      this.numSlots = numSlots;
      this.xa = xa;
      buffer = new double[numSlots];
      mutex = new BinarySemaphore(1);
      mutexP = new BinarySemaphore(1);
      mutexC = new BinarySemaphore(1);
      elements = new CountingSemaphore(0);
      spaces = new CountingSemaphore(numSlots);
      System.out.println("AnimatedBoundedBuffer alive, numSlots="
         + numSlots);
   }

   public void deposit(String name, double value) {
      P(spaces);
      P(mutexP);

// Move the producer to the slot.
      xa.moveTo(name, "Sc"+putIn);
// Change a full buffer slot's symbol to be solid black.
      xa.fill("slot"+putIn, xa.SOLID);

      buffer[putIn] = value;
      putIn = (putIn + 1) % numSlots;
      V(mutexP);
      P(mutex);
      count++;
      System.out.println(" after deposit, count=" + count
         + ", putIn=" + putIn);
      V(mutex);
      V(elements);
   }

   public double fetch(String name) {
      double value;
      P(elements);
      P(mutexC);

// Move the consumer to the slot.
      xa.moveTo(name, "Sc"+takeOut);
// Change an empty buffer slot's symbol to an outline black circle.
      xa.fill("slot"+takeOut, xa.OUTLINE);

      value = buffer[takeOut];
      takeOut = (takeOut + 1) % numSlots;
      V(mutexC);
      P(mutex);
      count--;
      System.out.println(" after fetch, count=" + count
         + ", takeOut=" + takeOut);
      V(mutex);
      V(spaces);
      return value;
   }
}

class Producer extends MyObject implements Runnable {

   private int pNap = 0; // milliseconds
   private AnimatedBoundedBuffer bb = null;
   private XtangoAnimator xa = null;

   public Producer(String name, int pNap, AnimatedBoundedBuffer bb,
         XtangoAnimator xa) {
      super(name);
      this.pNap = pNap;
      this.bb = bb;
      this.xa = xa;
   }

   public void run() {
      double item;
      int napping;
      while (true) {
         napping = 1 + (int) random(pNap);
         System.out.println("age=" + age() + ", " + getName()
            + " napping for " + napping + " ms");
         nap(napping);
         item = random();
         System.out.println("age=" + age() + ", " + getName()
            + " produced item " + item);

// Change a wants-to-deposit producers's symbol to be solid green.
         xa.color(getName(), Color.green);
         xa.fill(getName(), xa.SOLID);

         bb.deposit(getName(), item);

// Change producer's symbol to an outline black circle.
         xa.fill(getName(), xa.OUTLINE);
// Move the producer back to its original position.
         xa.moveTo(getName(), getName()+"c");
         xa.color(getName(), Color.black);

         System.out.println(getName() + " deposited item " + item);
      }
   }
}

class Consumer extends MyObject implements Runnable {

   private int cNap = 0; // milliseconds
   private AnimatedBoundedBuffer bb = null;
   private XtangoAnimator xa = null;

   public Consumer(String name, int cNap, AnimatedBoundedBuffer bb,
         XtangoAnimator xa) {
      super(name);
      this.cNap = cNap;
      this.bb = bb;
      this.xa = xa;
   }

   public void run() {
      double item;
      int napping;
      while (true) {
         napping = 1 + (int) random(cNap);
         System.out.println("age=" + age() + ", " + getName()
            + " napping for " + napping + " ms");
         nap(napping);
         System.out.println("age=" + age() + ", " + getName()
            + " wants to consume");

// Change a wants-to-fetch consumer's symbol to be solid blue.
         xa.color(getName(), Color.blue);
         xa.fill(getName(), xa.SOLID);

         item = bb.fetch(getName());

// Change consumers's symbol to an outline black circle.
         xa.fill(getName(), xa.OUTLINE);
// Move the consumer back to its original position.
         xa.moveTo(getName(), getName()+"c");
         xa.color(getName(), Color.black);

         System.out.println(getName() + " fetched item " + item);
      }
   }
}

class AnimatedBBPC extends MyObject {

   private static final double TWO_PI = 2.0 * Math.PI;

   public static void main(String[] args) {

      // parse command line arguments, if any, to override defaults
      GetOpt go = new GetOpt(args, "Us:P:C:p:c:R:");
      go.optErr = true;
      String usage = "Usage: -s numSlots -P numP -C numC"
                 + " -p pNap -c cNap -R runTime";
      int ch = -1;
      int numSlots = 20;
      int numProducers = 2;
      int numConsumers = 3;
      int pNap = 3;       // defaults
      int cNap = 2;       // in
      int runTime = 60;   // seconds
      while ((ch = go.getopt()) != go.optEOF) {
         if      ((char)ch == 'U') {
            System.out.println(usage);  System.exit(0);
         }
         else if ((char)ch == 's')
            numSlots = go.processArg(go.optArgGet(), numSlots);
         else if ((char)ch == 'P')
            numProducers = go.processArg(go.optArgGet(), numProducers);
         else if ((char)ch == 'C')
            numConsumers = go.processArg(go.optArgGet(), numConsumers);
         else if ((char)ch == 'p')
            pNap = go.processArg(go.optArgGet(), pNap);
         else if ((char)ch == 'c')
            cNap = go.processArg(go.optArgGet(), cNap);
         else if ((char)ch == 'R')
            runTime = go.processArg(go.optArgGet(), runTime);
         else {
            System.err.println(usage);  System.exit(1);
         }
      }
      System.out.println("AnimatedBBPC: numSlots=" + numSlots
         + ", numProducers=" + numProducers
         + ", numConsumers=" + numConsumers
         + ", pNap=" + pNap + ", cNap=" + cNap + ", runTime=" + runTime);

      XtangoAnimator xa = new XtangoAnimator();
      xa.begin();
// Change coordinates so 0,0 is the center, then create a big black outline
// circle to be the bounded buffer.
      xa.coords(-1.7f, -1.7f, 1.5f, 1.5f);
      xa.circle("buffer", 0.0f, 0.0f, 1.0f, Color.black, xa.OUTLINE);
// Put some annotated symbols on the screen.
      xa.circle("as1", -1.6f, -1.6f, 0.05f, Color.blue, xa.SOLID);
      xa.text("as2", -1.5f, -1.625f, false, Color.black,
         "consumer wants to fetch");
      xa.circle("as3", -1.6f, -1.4f, 0.05f, Color.green, xa.SOLID);
      xa.text("as4", -1.5f, -1.425f, false, Color.black,
         "producer wants to deposit");
      xa.circle("as5", -1.6f, -1.1f, 0.1f, Color.black, xa.OUTLINE);
      xa.text("as6", -1.4f, -1.125f, false, Color.black, "empty buffer slot");
      xa.circle("as7", -1.6f, -0.8f, 0.1f, Color.black, xa.SOLID);
      xa.text("as8", -1.4f, -0.825f, false, Color.black, "full buffer slot");
// Put some producers, small black outline circles, near the buffer.
// Also put a point for each producer to move back to after depositing.
      xa.text("as9", -0.5f+0.12f*(numProducers+1), -1.61f, false, Color.black,
         "producers");
      for (int i = 0; i < numProducers; i++) {
         xa.circle("PRODUCER"+i, -0.38f+0.12f*i, -1.6f, 0.05f,
            Color.black, xa.OUTLINE);
         xa.circle("PRODUCER"+i+"c", -0.38f+0.12f*i, -1.6f, 0.001f,
            Color.black, xa.OUTLINE);
      }
// Put some consumers, small black outline circles, near the buffer.
// Also put a point for each consumer to move back to after fetching.
      xa.text("as10", -0.5f+0.12f*(numConsumers+1), -1.41f, false, Color.black,
         "consumers");
      for (int i = 0; i < numConsumers; i++) {
         xa.circle("Consumer"+i, -0.38f+0.12f*i, -1.4f, 0.05f,
            Color.black, xa.OUTLINE);
         xa.circle("Consumer"+i+"c", -0.38f+0.12f*i, -1.4f, 0.001f,
            Color.black, xa.OUTLINE);
      }
// Put the buffer slots, black outline circles, around the buffer.
      for (int i = 0; i <numSlots; i++) {
         xa.circle("slot"+i, (float)Math.sin(i*(TWO_PI/numSlots)),
          (float)Math.cos(i*(TWO_PI/numSlots)), 0.1f, Color.black, xa.OUTLINE);
// Number the buffer slots.
         xa.text("Sn"+i, 0.85f*(float)Math.sin(i*(TWO_PI/numSlots)),
          0.85f*(float)Math.cos(i*(TWO_PI/numSlots)), true, Color.black, ""+i);
// Put nearly invisible circles (points) near each buffer slot to be places
// the producers and consumers can be moved to when they to deposit into or
// fetch from the buffer.
         xa.circle("Sc"+i, 1.15f*(float)Math.sin(i*(TWO_PI/numSlots)),
            1.15f*(float)Math.cos(i*(TWO_PI/numSlots)), 0.001f, Color.black,
            xa.OUTLINE);
      }

      // create the bounded buffer
      AnimatedBoundedBuffer bb = new AnimatedBoundedBuffer(numSlots, xa);

      // start the Producer and Consumer threads
      Thread[] producer = new Thread[numProducers];
      Thread[] consumer = new Thread[numConsumers];
      for (int i = 0; i < numProducers; i++)
         producer[i] = new Thread(
            new Producer("PRODUCER"+i, pNap*1000, bb, xa));
      for (int i = 0; i < numConsumers; i++)
         consumer[i] = new Thread(
            new Consumer("Consumer"+i, cNap*1000, bb, xa));
      for (int i = 0; i < numProducers; i++) producer[i].start();
      for (int i = 0; i < numConsumers; i++) consumer[i].start();
      System.out.println("All threads started");

      // let them run for a while
      nap(runTime*1000);
      System.out.println("age()=" + age()
         + ", time to stop the threads and exit");
      for (int i = 0; i < numProducers; i++) producer[i].stop();
      for (int i = 0; i < numConsumers; i++) consumer[i].stop();
      xa.end();
      System.exit(0);
   }
}
