import java.io.*;
import java.awt.*;
import Utilities.*;

class FibonacciFrame extends Frame {

   private TextArea textArea = null;
   private Fibonacci fibonacci = null;
   private Thread thread = null;

   public FibonacciFrame(String name) {
      super();
      setTitle("FibonacciFrame" + name);
      setLayout(new BorderLayout());
      Panel p = new Panel();
      p.setLayout(new FlowLayout());
      p.add(new Button("Start"));
      p.add(new Button("Suspend"));
      p.add(new Button("Resume"));
      p.add(new Button("Faster"));
      p.add(new Button("Slower"));
      p.add(new Button("Stop"));
      p.add(new Button("Close"));
      p.add(new Button("Quit"));
      add("North", p);
      textArea = new TextArea(12, 40);
      // the TextArea's memory may eventually
      // fill up and freeze the program
      textArea.setEditable(false);
      // change font, if desired
      // textArea.setFont(new Font("Helvetica", Font.PLAIN, 12));
      add("Center", textArea);
      fibonacci = new Fibonacci(textArea);
   }

   public boolean handleEvent(Event evt) {  // recommended for Windows 95
      if (evt.id == Event.WINDOW_DESTROY) { // by "Core Java" pages 181-182
         System.exit(0);
      } else return super.handleEvent(evt);
      return true;
   }

   public boolean action(Event evt, Object arg) {
      int newNap;
      if (evt.target instanceof Button) {
         if (arg.equals("Start")) {
            if (thread != null) thread.stop();
            thread = new Thread(fibonacci);
            // without lowering the thread's priority,
            // a newNap < 32 has sometimes frozen
            // the buttons
            thread.setPriority(Thread.MIN_PRIORITY);
            thread.start();
         } else if (arg.equals("Suspend")) {
            if (thread != null) thread.suspend();
         } else if (arg.equals("Resume")) {
            if (thread != null) thread.resume();
         } else if (arg.equals("Faster")) {
            if (thread != null) {
               thread.suspend();
               newNap = fibonacci.napping;
               newNap /= 2;
               newNap = newNap < 2 ? 2 : newNap;
               fibonacci.napping = newNap;
               textArea.appendText("napping=" + newNap + "\n");
               thread.resume();
            }
         } else if (arg.equals("Slower")) {
            if (thread != null) {
               thread.suspend();
               newNap = fibonacci.napping;
               newNap *= 2;
               fibonacci.napping = newNap;
               textArea.appendText("napping=" + newNap + "\n");
               thread.resume();
            }
         } else if (arg.equals("Stop")) {
            if (thread != null) thread.stop();
         } else if (arg.equals("Close")) {
            if (thread != null) thread.stop();
            this.hide();
            this.dispose();
         } else if (arg.equals("Quit")) {
            System.exit(0);
         } else return super.action(evt, arg);
      } else return super.action(evt, arg);
      return true;
   }

   public static void main(String[] args) {
      Frame frame = null;
      frame = new FibonacciFrame(" one");
      frame.pack();
      frame.show();
      frame = new FibonacciFrame(" two");
      frame.pack();
      frame.show();
   }
}

class Fibonacci extends MyObject implements Runnable {

   private TextArea textArea = null;
   public int napping = 256;         // public so Frame can modify it

   public Fibonacci(TextArea ta) {
      super();
      textArea = ta;
   }

   public void run() {
      long x = (long)random(1,100);
      long y = (long)random(1,100);
      long z;
      while (true) {
         z = x + y;
         textArea.appendText("x=" + x + ", y=" + y
            + ", z=" + z + "\n");
         nap(napping);
         x = y;
         y = z;
      }
   }
}
