import Utilities.*;

class BufferItem {
   public volatile double value = 0;          // multiple threads acccess
   public volatile boolean occupied = false;  // so make these `volatile'
}

class BoundedBuffer {

   private int numSlots = 0;
   private BufferItem[] buffer = null;
   private int putIn = 0, takeOut = 0;
   // private int count = 0;

   public BoundedBuffer(int numSlots) {
      this.numSlots = numSlots;
      buffer = new BufferItem[numSlots];
      for (int i = 0; i < numSlots; i++) {
         buffer[i] = new BufferItem();
      }
      System.out.println("BoundedBuffer alive, numSlots="
         + numSlots);
   }

   public void deposit(double value) {
      while (buffer[putIn].occupied) /* busy wait */ ;
      buffer[putIn].value = value;                              // A
      buffer[putIn].occupied = true;                            // B
      putIn = (putIn + 1) % numSlots;
      // count++;                                       // race condition!!!
      System.out.println(" after deposit putIn=" + putIn);
   }

   public double fetch() {
      double value;
      while (!buffer[takeOut].occupied) /* busy wait */ ;
      value = buffer[takeOut].value;                            // C
      buffer[takeOut].occupied = false;                         // D
      takeOut = (takeOut + 1) % numSlots;
      // count--;                                       // race condition!!!
      System.out.println(" after fetch takeOut=" + takeOut);
      return value;
   }
}

/* ............... Example compile and run(s)

D:\Threads>javac bwbb.java bbpc.java

D:\Threads>java ProducerConsumer -s 6 -p 2 -c 2 -R 10
ProducerConsumer: numSlots=6, pNap=2, cNap=2, runTime=10
BoundedBuffer alive, numSlots=6
All threads started
age=50, PRODUCER napping for 997 ms
age=110, Consumer napping for 1863 ms
age=1150, PRODUCER produced item 0.196652
 after deposit putIn=1
PRODUCER deposited item 0.196652
age=1150, PRODUCER napping for 1091 ms
age=2030, Consumer wants to consume
 after fetch takeOut=1
Consumer fetched item 0.196652
age=2030, Consumer napping for 439 ms
age=2250, PRODUCER produced item 0.268426
 after deposit putIn=2
PRODUCER deposited item 0.268426
age=2250, PRODUCER napping for 362 ms
age=2470, Consumer wants to consume
 after fetch takeOut=2
Consumer fetched item 0.268426
age=2470, Consumer napping for 946 ms
age=2630, PRODUCER produced item 0.958945
 after deposit putIn=3
PRODUCER deposited item 0.958945
age=2630, PRODUCER napping for 1939 ms
age=3400, Consumer wants to consume
 after fetch takeOut=3
Consumer fetched item 0.958945
age=3400, Consumer napping for 1312 ms
age=4610, PRODUCER produced item 0.078194
 after deposit putIn=4
PRODUCER deposited item 0.078194
age=4610, PRODUCER napping for 619 ms
age=4720, Consumer wants to consume
 after fetch takeOut=4
Consumer fetched item 0.078194
age=4720, Consumer napping for 576 ms
age=5220, PRODUCER produced item 0.0630488
 after deposit putIn=5
PRODUCER deposited item 0.0630488
age=5220, PRODUCER napping for 934 ms
age=5320, Consumer wants to consume
 after fetch takeOut=5
Consumer fetched item 0.0630488
age=5320, Consumer napping for 1993 ms
age=6150, PRODUCER produced item 0.580876
 after deposit putIn=0
PRODUCER deposited item 0.580876
age=6150, PRODUCER napping for 851 ms
age=7030, PRODUCER produced item 0.974548
 after deposit putIn=1
PRODUCER deposited item 0.974548
age=7030, PRODUCER napping for 1729 ms
age=7300, Consumer wants to consume
 after fetch takeOut=0
Consumer fetched item 0.580876
age=7300, Consumer napping for 496 ms
age=7850, Consumer wants to consume
 after fetch takeOut=1
Consumer fetched item 0.974548
age=7850, Consumer napping for 1480 ms
age=8730, PRODUCER produced item 0.0263791
 after deposit putIn=2
PRODUCER deposited item 0.0263791
age=8730, PRODUCER napping for 1249 ms
age=9330, Consumer wants to consume
 after fetch takeOut=2
Consumer fetched item 0.0263791
age=9330, Consumer napping for 1890 ms
age=9990, PRODUCER produced item 0.948153
 after deposit putIn=3
PRODUCER deposited item 0.948153
age=9990, PRODUCER napping for 1377 ms
age()=10050, time to stop the threads and exit
                                            ... end of example run(s)  */
