import Utilities.*;

class Racer extends MyObject implements Runnable {

   private int M = 0;
   private volatile int sum = 0;

   public Racer(String name, int M) {
      super(name);
      this.M = M;
      System.out.println("age()=" + age() + ", "
         + getThreadName() + " is alive, M=" + M);
   }

   private int fn(int j, int k) {
      int total = j;
      for (int i = 1;  i <= k; i++) total += i;
      return total;
   }

   public void run() {
      System.out.println("age()=" + age() + ", "
         + getThreadName() + " is running");
      for (int m = 1; m <= M; m++)
/*
 * "N = N + 1" type lost update (race condition) in following line
 */
         sum = fn(sum, m);

      System.out.println("age()=" + age() + ", "
         + getThreadName() + " is done, sum = " + sum);
   }
}

class RaceTwoThreads extends MyObject {

   private static int M = 10;
   private final static int numRacers = 2;

   public static void main(String[] args) {

      // parse command line arguments, if any, to override defaults
      GetOpt go = new GetOpt(args, "UM:");
      go.optErr = true;
      String usage = "Usage: -M m";
      int ch = -1;
      while ((ch = go.getopt()) != go.optEOF) {
         if      ((char)ch == 'U') {
            System.out.println(usage);  System.exit(0);
         }
         else if ((char)ch == 'M')
            M = go.processArg(go.optArgGet(), M);
         else {
            System.err.println(usage);  System.exit(1);
         }
      }
      System.out.println("RaceTwoThreads: M=" + M);

      // start the two threads, both in the same object
      // so they share one instance of its variable sum
      Racer racerObject = new Racer("Racer", M);
      Thread[] racer = new Thread[numRacers];
      for (int i = 0; i < numRacers; i++)
         racer[i] = new Thread(racerObject, "Racer" + i);
      for (int i = 0; i < numRacers; i++) racer[i].start();
      System.out.println("age()=" + age() +
         ", all Racer threads started");

      // wait for them to finish
      try {
         for (int i = 0; i < numRacers; i++) racer[i].join();
      } catch (InterruptedException e) {
         System.err.println("interrupted out of join");
      }

      // correct race-free final value of sum is 2*220 = 440 for M of 10
      // and 2*167167000 = 334334000 for M of 1000
      System.out.println("RaceTwoThreads done");
      System.exit(0);
   }
}

/* ............... Example compile and run(s)

D:\Threads>javac race.java

D:\Threads>java RaceTwoThreads
RaceTwoThreads: M=10
age()=50, Racer0 is alive
age()=50, Racer1 is alive
age()=50, all Racer threads started
age()=50, Racer1 is running
age()=110, Racer1 is done, sum = 220
age()=50, Racer0 is running
age()=110, Racer0 is done, sum = 440
RaceTwoThreads: sum=440

D:\Threads>java RaceTwoThreads -M 1000
RaceTwoThreads: M=1000
age()=50, Racer0 is alive
age()=50, Racer1 is alive
age()=50, all Racer threads started
age()=50, Racer1 is running
age()=50, Racer0 is running
age()=1480, Racer1 is done, sum = 167167000
age()=1810, Racer0 is done, sum = 169329940
RaceTwoThreads: sum=169329940
                                            ... end of example run(s)  */
