package Synchronization;

import Utilities.*;

// a Barrier for n threads: all must arrive before any are released
public class Barrier extends MyObject implements Runnable {

   private int n = -1;
   private CountingSemaphore arrive = null;
   private BinarySemaphore[] release = null;
   private Thread thread = null;

   public Barrier(int n) {
      super("Barrier for " + n + " threads");
      this.n = n;
      arrive = new CountingSemaphore(0);
      release = new BinarySemaphore[n];
      for (int i = 0; i < n; i++) release[i] = new BinarySemaphore(0);
      thread = new Thread(this);
      thread.start();
   }

   public String toString() {
      return getName() + ", arrive=" + arrive;
   }

   public void barrier(int i) {
      V(arrive);
      P(release[i]);
   }

   public void stop() {
      thread.stop();
   }

   public void run() {
      while (true) {
         // If we are worried about two different threads mistakenly
         // calling barrier(i) for the same i, we can make arrive into
         // an array of semaphores like semaphore release already is.
         for (int i = 0; i < n; i++) P(arrive);
         // Second semaphore must be an array to avoid race conditions
         // due to arbitrary context switches and relative CPU speeds.
         for (int i = 0; i < n; i++) V(release[i]);
      }
   }
}
